# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2017-10-26 09:59:52.871866
from __future__ import unicode_literals
import datetime

version = "2.0"
githash = "ed2e7897862ae9e979b336df670d7a5541cdd8f0"


major = 2
minor = 0
bugfix = 0

release = True
timestamp = datetime.datetime(2017, 10, 26, 9, 59, 52, 871866)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
