# Licensed under a 3-clause BSD style license - see LICENSE.rst

def get_package_data():
    return {
        _ASTROPY_PACKAGE_NAME_ + '.data_factories': ['configurations/*.yaml', 'ifucube_popup.ui']}
